{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Model Server"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The Intel Neural Compute Stick 2 is a compact and yet powerful Vision Processing Unit (VPU) which is optimized to perform the complex matrix calculations required for convolutional neural networks. Similar to a GPU, the VPU is engineered specifically for computer vision inference on images and videos.\n",
    "\n",
    "The VPU is initalized with the network structure of the neural networks as well as the weights; these networks are then allocated space on the device. Total time to initialize a network is approximately 30 seconds, however once it has been initalized successive inference on a loaded model is processed in milliseconds. \n",
    "\n",
    "In order for multiple modules to make use of the VPU, it is therefor desirable for the device to be initialized once with multiple models. \n",
    "\n",
    "To accomplish this I have engineered the Model Server. The Model Server is responsable for initializing the VPU device, and registering the various Networks with the device. The Model Server then makes an Inference Service API available to the modules. If a module needs inference performed, it uses the Model Server API to provide the server with the kind of model needed, as well as the image requirements. The model server then uses its most recent image and performs inference on the VPU, returning the results to the model module where additional  processing can take place.\n",
    "\n",
    "In order to receive the most recent images, the Modeling Server offers a second API endpoint for the Image Server Module. A limitation of ROS Noetic is that a service node cannot subscribe to other nodes like normal nodes can; this is because the service node needs to continually monitor the API requests. As such, the Model Server cannot obtain images on its own outside of an API call. In order to minimize the bandwidth and the duplication of images, the Image Server offers a kind of \"reverse\" service. Rather than acting as an API endpoint, the Image Server makes a service call to the Model Server requesting to give it an image via the Image Fetching Service offered by the model server. The Model Server then accepts that image, and the Image Server carries on with the management of the Image flags and requests for the nodes. "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "from viper_toolkit import Dissect\n",
    "import inspect"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## The Model Server Module"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "metadata": {
    "scrolled": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "class ModelServer(object):\n",
      "    \n",
      "    def __init__(self):\n",
      "        self.setup_ros()\n",
      "        self.image = Image()\n",
      "        self.setup_inference_engine()\n",
      "        self.setup_models()\n",
      "        self.loop()\n",
      "        \n",
      "    def setup_ros(self):\n",
      "        rospy.init_node('model_server', log_level=rospy.DEBUG)\n",
      "        self.setup_parameters()\n",
      "        self.logger.i('Model Server Initialized')\n",
      "    \n",
      "    def setup_parameters(self):\n",
      "        # Instantiates the Name Manager and Logger to standardize the \n",
      "        # nodes name and abreviation throughout the addon modules\n",
      "        self.name = NameManager()\n",
      "        self.logger = Logger(self.name)\n",
      "        \n",
      "        # Creates an instance of the organizer I will be using for \n",
      "        # my parameters. This allows for dynamic updatign accross nodes\n",
      "        self.parameters = ParameterManager()\n",
      "        \n",
      "        # Sets this module to allow for dynamic updating\n",
      "        self.parameters.add(\n",
      "            Parameter(\n",
      "                name=\"updates\", \n",
      "                target=f\"{self.name.name}/dynamic\",\n",
      "                default=True, \n",
      "                dynamic=True))\n",
      "                \n",
      "        # Searches for the code of the Pode Detection Module. If the \n",
      "        # the module cannot be found, the default of PDM will be used\n",
      "        # and will be updated if the modules code changes when it comes\n",
      "        # online. This code is used to tell the inference engine which\n",
      "        # neural network to use for inference.\n",
      "        self.parameters.add(\n",
      "            Parameter(\n",
      "                name=\"PoseDetection_abv\", \n",
      "                target=rospy.search_param(\n",
      "                    'pose_detection_module/abv'), \n",
      "                default=\"PDM\", \n",
      "                dynamic=True))\n",
      "        \n",
      "        # Searches for the code of the Scene Segmentation module. Will\n",
      "        # use default value if not found, and will update dynamically \n",
      "        # if the code later becomes available.\n",
      "        self.parameters.add(\n",
      "            Parameter(\n",
      "                name = \"SceneSegmentation_abv\", \n",
      "                target = rospy.search_param(\n",
      "                    'scene_segmentation_module/abv'), \n",
      "                default = \"SEG\", \n",
      "                dynamic = True))\n",
      "\n",
      "    def setup_inference_engine(self):\n",
      "        \n",
      "        # Sets up the OpenVino Inference Engine core to handle\n",
      "        # multiple models\n",
      "        self.ie = IECore()\n",
      "        self.logger.i(\"IECore Online\")\n",
      "\n",
      "    def setup_models(self):\n",
      "        \n",
      "        # Create a dictionary of models. The specific model will be\n",
      "        # be identified at the time of inference.\n",
      "        self.models = {}\n",
      "        \n",
      "        # Initializes the models. All models will be loaded to the \n",
      "        # Neural Compute Stick Vision Processing Unit.\n",
      "        self.setup_pose_detector(device=\"MYRIAD\")\n",
      "        self.setup_scene_segmentor(device=\"MYRIAD\")\n",
      "\n",
      "    def setup_pose_detector(self, device):\n",
      "        \n",
      "        # Initializes a viper model pointing to the Artificial Neural\n",
      "        # Network Architecture and the weights.\n",
      "        pose_detector = ViperModel(\n",
      "            package_name = 'pose_detection_module',\n",
      "            model_xml = 'model/human-pose-estimation-0001.xml',\n",
      "            weights_bin = 'model/human-pose-estimation-0001.bin'\n",
      "            )\n",
      "        \n",
      "        # These model paramters are then loaded into the Intel\n",
      "        # Nerual Compute Stick inference engine. \n",
      "        self.pose_engine = PoseDetectionModel(\n",
      "            ie = self.ie,\n",
      "            viper_model = pose_detector,\n",
      "            device = device,\n",
      "            model_tag = \"PD.\",\n",
      "            model_name = \"PoseDetectionModel\"\n",
      "            )\n",
      "        \n",
      "        # The path to the inference function of the model is \n",
      "        # saved in the model dictionary. This will be called when\n",
      "        # a node makes a service API request via its access point.\n",
      "        \n",
      "        #self.models[f'{self.parameters.PoseDetection_abv}'] \\\n",
      "        #    = self.pose_engine.run_pose_estimation\n",
      "            \n",
      "        self.models[\"POS\"] = self.pose_engine.run_pose_estimation\n",
      "        self.logger.i('Pose Detection Model Loaded')\n",
      "\n",
      "    def setup_scene_segmentor(self, device):\n",
      "        \n",
      "        # Here we set up the Scene Segmentation model.\n",
      "        scene_segmentor = ViperModel(\n",
      "            package_name = \"scene_segmentation_module\",\n",
      "            model_xml = 'model/road-segmentation-adas-0001.xml',\n",
      "            weights_bin = 'model/road-segmentation-adas-0001.bin'\n",
      "            )\n",
      "            \n",
      "        # We load the model to the same device, which had been \n",
      "        # already been initialized at the server startup. This allows\n",
      "        # us to add and remove models as needed, without shutting\n",
      "        # the VPU down, while also sharing the device.\n",
      "        self.scene_segmentor = SceneSegmentationModel(\n",
      "            ie = self.ie,\n",
      "            viper_model = scene_segmentor,\n",
      "            device = device,\n",
      "            model_tag = \"SD.\",\n",
      "            model_name = \"SceneSegmentationModel\"\n",
      "            )\n",
      "        \n",
      "        # We load the Scene Segmentation model to the model dictionary.\n",
      "        \n",
      "        #self.models[f'{self.parameters.SceneSegmentation_abv}'] \\\n",
      "        #    = self.scene_segmentor.run_scene_segmentation\n",
      "        \n",
      "        self.models[\"SEG\"] = self.scene_segmentor.run_scene_segmentation\n",
      "        self.logger.i(\"Scene Segmentor Loaded\")\n",
      "\n",
      "    def handle_model_rqst(self, request):\n",
      "        \n",
      "        # This function is the API callback. Upon receiving a proper\n",
      "        # request via the service endpoint, the requested model\n",
      "        # is saved as 'model'\n",
      "        \n",
      "        self.process_timer.lap(\"Modeling Service\")\n",
      "        self.logger.i(\"Modeling Request Received\")\n",
      "        model = request.model.data\n",
      "        print(model)\n",
      "        # The image which we will be performing inference on is \n",
      "        # is converted from a ROS message to a  numpy image array.\n",
      "        try:\n",
      "            bridge = CvBridge()\n",
      "            image = bridge.imgmsg_to_cv2(\n",
      "                self.image,\n",
      "                desired_encoding=\"rgb8\"\n",
      "                )\n",
      "                \n",
      "            #self.logger.i(f'Image reshaped to: {np.shape(image)}')\n",
      "            \n",
      "        except CvBridgeError as cv_error:\n",
      "            print(\"ERRRRRRERERRROROR\")\n",
      "            self.logger.e(f'\"CvBridge Error: {cv_error}')\n",
      "            \n",
      "            print(cv_error)\n",
      "            \n",
      "            #If this process fails we will need to return to try \n",
      "            # on the next loop. This sometimes happens due to \n",
      "            # packet loss.\n",
      "            \n",
      "            return\n",
      "            \n",
      "        # We use the resulting image to perform inference.\n",
      "        self.process_timer.lap(model)\n",
      "        \n",
      "        try:\n",
      "            results = self.models[model](frame=image)\n",
      "            print(results)\n",
      "        except:\n",
      "            self.logger.e(f\"Error modeling {model}\")\n",
      "            return\n",
      "            \n",
      "        self.process_timer.time(model)\n",
      "        self.process_timer.lap(f'Flatten {model}')\n",
      "        \n",
      "        # We need to serialize all of the inferences accross models and\n",
      "        # accross results in order to format them in a standard\n",
      "        # and consistent message packet.\n",
      "        try:\n",
      "            packet = self.flatten_inference(results)\n",
      "        except:\n",
      "            self.logger.e(f\"Error flattening {model}\")\n",
      "            \n",
      "            return\n",
      "            \n",
      "        self.process_timer.time(f'Flatten {model}')\n",
      "        self.process_timer.time(\"Modeling Service\")\n",
      "        \n",
      "        # Upon success we will return the message packet to the API for\n",
      "        # delivery to the client.\n",
      "        \n",
      "        return packet\n",
      "\n",
      "    def handle_fetch_rqst(self, request):\n",
      "        \n",
      "        # This is the Service API callback for the Image Server Service.\n",
      "        # Upon receiving a request from the model server to deliver\n",
      "        # an image, this process saves that image message (self.image).\n",
      "        #self.process_timer.lap(\"Image Service\")\n",
      "        #self.logger.i(\"Image request received\")\n",
      "        \n",
      "        try:\n",
      "            self.image = request.image\n",
      "            receipt = True\n",
      "        except:\n",
      "            receipt = False\n",
      "            \n",
      "        #self.process_timer.time(\"Image Service\")\n",
      "        # We then respond with a message indicating the outcome of the\n",
      "        # request.\n",
      "        \n",
      "        return ImageRequestResponse(Bool(receipt))\n",
      "\n",
      "    def flatten_inference(self, results_array):\n",
      "        \n",
      "        # This function converts our results array into a \n",
      "        # [std_msgs.msg/int32] array containing the ndarry shape and \n",
      "        # a [std_msgs.msg/float32] array containing the flattened array.\n",
      "        \n",
      "        # Load the InferenceResults Template\n",
      "        packet = InferenceResults()\n",
      "        print(results_array)\n",
      "        try:\n",
      "            structure = results_array.shape\n",
      "            rospy.logdebug(\"[MOD] mask structure: %s\",structure)\n",
      "            structure = list(structure) \n",
      "            structure = np.array(structure).astype(np.int32)\n",
      "            structure = list(structure)\n",
      "            rospy.logdebug(\"[MOD] Converted params: %s\", structure)\n",
      "            packet.structure = structure # [Int32]\n",
      "            rospy.logdebug(\"[MOD] Structure sent: %s\",packet.structure)\n",
      "            \n",
      "        except:\n",
      "            rospy.logerr(\"[MOD] Cannot save mask structure.\")\n",
      "            \n",
      "        try:\n",
      "            inferences = np.array(list(results_array.flatten()))\n",
      "            packet.inferences = inferences # [float32]\n",
      "            \n",
      "        except:\n",
      "            rospy.logerr(\"[MOD] Cannot save Inference shape.\")\n",
      "            rospy.logerr(\"[MOD] Inference shape is: %s\", inferences.shape)\n",
      "        print(packet)\n",
      "        return packet\n",
      "\n",
      "    def loop(self):\n",
      "        \n",
      "        while not rospy.is_shutdown(): \n",
      "            \n",
      "            # This function will keep our parameters dynamically updated\n",
      "            # in case that they change as other nodes come online\n",
      "            # or offile. Parameters marked as \"updates = True\" will\n",
      "            # update accordingly. \n",
      "            \n",
      "            if self.parameters.updates == True:\n",
      "                    self.parameters.update()\n",
      "            \n",
      "            # This function Initializes my custom process timer module, \n",
      "            # which computes the time between any two proceses by name.\n",
      "            # It also connects with my custom \"Logger\" module, the \n",
      "            # \"Parameter\" module, and my \"NameManager\" module to help me\n",
      "            # track information accross Nodes.\n",
      "            \n",
      "            self.process_timer = ProcessTimer(logger = self.logger)\n",
      "            \n",
      "            # This is the API access point for the Image Server. The\n",
      "            # client submits an Image Request message, which contains\n",
      "            # an ROS Image message. In return they will receive a Bool\n",
      "            # indicating the outcome of the request.\n",
      "            \n",
      "            self.image_fetching_service = rospy.Service(\n",
      "                'model_server/image_fetching_service', \n",
      "                ImageRequest, \n",
      "                self.handle_fetch_rqst\n",
      "                )\n",
      "            #self.logger.i(\"Image Fetching Service Online\")\n",
      "            \n",
      "            # This is the API access point for the Modeling Service. The\n",
      "            # client submits a three letter model code, and the service\n",
      "            # will use the most recent image to perform inference. It\n",
      "            # then returns two arrays; one array 'Structure' contains \n",
      "            # the shape of the original inference output, and a second\n",
      "            # array \"Inferences\" contains the flattened 1-Dimnesional \n",
      "            # array.\n",
      "            \n",
      "            self.modeling_service = rospy.Service(\n",
      "                'model_server/modeling_service', \n",
      "                ModelRequest, \n",
      "                self.handle_model_rqst\n",
      "                )\n",
      "                \n",
      "            #self.logger.i(\"Modeling Service Online\")\n",
      "\n",
      "            #self.logger.i(\"Model Server Online\")\n",
      "            \n",
      "            # Rospy.spin() keeps our kernal active during times of \n",
      "            # inactivity.\n",
      "            \n",
      "            rospy.spin()\n",
      "\n"
     ]
    }
   ],
   "source": [
    "from scripts import modelserver\n",
    "from scripts.modelserver import ModelServer\n",
    "source = inspect.getsource(ModelServer)\n",
    "print (source)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [],
   "source": [
    "from model_server.srv import ImageRequest, ImageRequestResponse"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 20,
   "metadata": {},
   "outputs": [],
   "source": [
    "var = ImageRequest"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 24,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "status: \n",
       "  data: False"
      ]
     },
     "execution_count": 24,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "ImageRequestResponse()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 25,
   "metadata": {},
   "outputs": [],
   "source": [
    "from model_server.msg import InferenceResults"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 32,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "structure: []\n",
      "inferences: []\n",
      "Structure Datatype: <class 'list'>\n",
      "Inferences Datatype: <class 'list'>\n"
     ]
    }
   ],
   "source": [
    "print(InferenceResults())\n",
    "print (f'Structure Datatype: {type(InferenceResults().structure)}')\n",
    "print (f'Inferences Datatype: {type(InferenceResults().inferences)}')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.10"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
